import argparse
import gzip
import io
import os
from pathlib import Path

# Extensions that it's not worth trying to compress
SKIP_COMPRESS_EXTENSIONS = (
    # Images
    ".jpg",
    ".jpeg",
    ".png",
    ".gif",
    ".webp",
    # Compressed files
    ".zip",
    ".gz",
    ".tgz",
    ".bz2",
    ".tbz",
    ".xz",
    ".br",
    # Flash
    ".swf",
    ".flv",
    # Fonts
    ".woff",
    ".woff2",
    # Video
    ".3gp",
    ".3gpp",
    ".asf",
    ".avi",
    ".m4v",
    ".mov",
    ".mp4",
    ".mpeg",
    ".mpg",
    ".webm",
    ".wmv",
)


def try_gzip(path):
    orig_bytes = path.read_bytes()
    with io.BytesIO() as f:
        with gzip.GzipFile(
            filename="", mode="wb", fileobj=f, compresslevel=9, mtime=0
        ) as gz_file:
            gz_file.write(orig_bytes)
        gz_bytes = f.getvalue()
        orig_len = len(orig_bytes)
        gz_len = len(gz_bytes)
        if gz_len < orig_len * 0.9:
            print(
                f"{path!s} has been shrinked by {orig_len - gz_len} bytes to {int(100 * len(gz_bytes) / len(orig_bytes))}%"
            )
            Path(str(path) + ".gz").write_bytes(gz_bytes)
        else:
            print(f"{path!s} has been skipped because of missing gains")


def compress(root):
    for root, _dirs, files in os.walk(args.root):
        dir = Path(root)
        for filename in files:
            path = dir / filename
            if path.suffix in SKIP_COMPRESS_EXTENSIONS:
                # print(f"Skipping {str(path)} because of extension")
                continue

            try_gzip(path)

    return 0


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("root", help="Path containing static files to compress")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    raise SystemExit(compress(args.root))



# SPDX-FileCopyrightText: 2024-present Matthias Kestenholz <mk@feinheit.ch>
#
# SPDX-License-Identifier: MIT

try:
    from blacknoise._impl import BlackNoise  # noqa
except ModuleNotFoundError:
    pass



import os

from starlette.responses import FileResponse, PlainTextResponse

# Ten years is what nginx sets a max age if you use 'expires max;'
# so we'll follow its lead
FOREVER = f"max-age={10 * 365 * 24 * 60 * 60}, public, immutable"
A_LITTE_WHILE = "max-age=60, public"

_compress_content_encodings = {".gz": "gzip"}


class BlackNoise:
    def __init__(self, application, *, immutable_file_test=lambda *_a: False):
        self._files = {}
        self._prefixes = ()
        self._application = application

        self._immutable_file_test = immutable_file_test

    def add(self, path, prefix):
        self._prefixes = (*self._prefixes, prefix)

        for base, _dirs, files in os.walk(path):
            path_prefix = os.path.join(prefix, base[len(str(path)) :].strip("/"))
            self._files |= {
                os.path.join(path_prefix, file): os.path.join(base, file)
                for file in files
                if all(
                    # File is not a compressed file
                    not file.endswith(suffix)
                    # The uncompressed variant does not exist
                    or file.removesuffix(suffix) not in files
                    for suffix in _compress_content_encodings
                )
            }

    async def __call__(self, scope, receive, send):
        path = os.path.normpath(scope["path"].removeprefix(scope["root_path"]))

        if not path.startswith(self._prefixes):
            response = self._application

        elif scope["type"] != "http" or scope["method"] not in ("GET", "HEAD"):
            response = PlainTextResponse("Method Not Allowed", status_code=405)

        elif file := self._files.get(path):
            headers = {
                "access-control-allow-origin": "*",
                "cache-control": (
                    FOREVER if self._immutable_file_test(path) else A_LITTE_WHILE
                ),
            }
            gz_file = f"{file}.gz"
            if os.path.exists(gz_file):
                response = FileResponse(
                    gz_file, headers=headers | {"content-encoding": "gzip"}
                )
            elif encoding := next(
                (
                    encoding
                    for suffix, encoding in _compress_content_encodings.items()
                    if file.endswith(suffix)
                ),
                None,
            ):
                response = FileResponse(
                    file, headers=headers | {"content-encoding": encoding}
                )
            else:
                response = FileResponse(file, headers=headers)

        else:
            response = PlainTextResponse("Not Found", status_code=404)

        await response(scope, receive, send)



from pathlib import Path

import httpx
import pytest
from starlette.responses import PlainTextResponse

from blacknoise._impl import BlackNoise


async def app(scope, receive, send):
    response = PlainTextResponse(f"Hello from {scope['path']}")
    await response(scope, receive, send)


@pytest.fixture
def bn():
    this = Path(__file__).parent
    blacknoise = BlackNoise(app)
    blacknoise.add(this / "static", "/static/")
    return blacknoise


def test_files_contents(bn):
    this = Path(__file__).parent

    assert "/static/hello.txt" in bn._files
    assert "/static/hello.txt.gz" not in bn._files
    assert "/static/hello2.txt.gz" in bn._files
    assert "/static/hello3.txt" in bn._files
    assert "/static/foo" not in bn._files

    assert bn._files["/static/hello.txt"] == str(this / "static" / "hello.txt")


@pytest.mark.asyncio
async def test_static_file_serving(bn):
    transport = httpx.ASGITransport(app=bn)
    async with httpx.AsyncClient(
        transport=transport, base_url="http://testserver"
    ) as client:
        r = await client.get("/static/hello.txt")
        assert r.status_code == 200
        assert r.text == "world\n"
        assert r.headers["content-encoding"] == "gzip"

        r = await client.get("/static/hello2.txt.gz")
        assert r.status_code == 200
        assert r.text == "world2\n"
        assert r.headers["content-encoding"] == "gzip"

        r = await client.get("/static/hello3.txt")
        assert r.status_code == 200
        assert r.text == "world3\n"
        assert "content-encoding" not in r.headers

        r = await client.post("/static/hello.txt")
        assert r.status_code == 405

        r = await client.get("/static/foo")
        assert r.status_code == 404

        r = await client.get("/blub")
        assert r.status_code == 200
        assert r.text == "Hello from /blub"
